import base64
import json
from dataclasses import dataclass, field
from enum import Enum
from typing_extensions import Type, get_origin
from typing import Callable, Union, ForwardRef
from typing import TypeVar, Optional, Dict, List, Any
from urllib.parse import urljoin, quote_plus

import requests
from dataclasses_json import dataclass_json
from requests.exceptions import HTTPError
from requests.models import Response,RequestField, encode_multipart_formdata

from dataclasses import dataclass
from typing import BinaryIO, List, Optional, TypeVar, Union
import mimetypes

from servicestack.types import IDeleteDb, IPatchDb, IReturn, IReturnVoid, IGet, IPost, IPut, IPatch, \
    IDelete, IUpdateDb, QueryBase, ResponseStatus, EmptyResponse, GetAccessToken, GetAccessTokenResponse
from servicestack.log import Log
from servicestack.reflection import TypeConverters, to_dict, nameof, is_list, is_dict, _resolve_forwardref, \
    has_type_vars, _dict_with_string_keys, _get_type_vars_map, from_json, to_json
from servicestack.utils import ex_message, clean_camelcase


JSON_MIME_TYPE = "application/json"
AUTHORIZATION_HEADER = "Authorization"
CONTENT_TYPE = "Content-Type"
SS_TOKEN_COOKIE = "ss-tok"
SS_REFRESH_TOKEN_COOKIE = "ss-reftok"


def _resolve_response_type(request):
    t = type(request)

    def resolve_response_type():
        if isinstance(request, IReturn):
            for cls in t.__orig_bases__:
                if get_origin(cls) == IReturn and hasattr(cls, '__args__'):
                    candidate = cls.__args__[0]
                    if type(candidate) == ForwardRef:
                        return _resolve_forwardref(candidate, type(request))
                    return candidate
        if isinstance(request, IReturnVoid):
            return type(None)
        return None

    if hasattr(t, 'response_type'):
        ret = t.response_type()
        if has_type_vars(ret):
            # avoid reifying type vars if request type has concrete type return marker
            ret_candidate = resolve_response_type()
            if not has_type_vars(ret_candidate):
                return ret_candidate

            type_map = _dict_with_string_keys(_get_type_vars_map(t))
            if isinstance(ret, TypeVar):
                return get_origin(ret).__class_getitem__(type_map[f"{ret}"])
            reified_args = [type_map[f"{x}"] for x in ret.__args__]
            reified_type = get_origin(ret).__class_getitem__(*reified_args)
            return reified_type
        return ret
    else:
        return resolve_response_type()


def resolve_httpmethod(request):
    if isinstance(request, IGet) or isinstance(request, QueryBase):
        return "GET"
    if isinstance(request, IPost):
        return "POST"
    if isinstance(request, IPut) or isinstance(request, IUpdateDb):
        return "PUT"
    if isinstance(request, IPatch) or isinstance(request, IPatchDb):
        return "PATCH"
    if isinstance(request, IDelete) or isinstance(request, IDeleteDb):
        return "DELETE"
    return "POST"


def qsvalue(arg):
    if arg is None:
        return ""
    arg_type = type(arg)
    if arg_type == bool:
        return f"{arg}".lower()
    if is_list(arg_type):
        return "[" + ','.join([qsvalue(x) for x in arg]) + "]"
    if is_dict(arg_type):
        return "{" + ','.join([k + ":" + qsvalue(v) for k, v in arg]) + "}"
    if arg_type is str:
        return quote_plus(arg)
    if arg_type in TypeConverters.serializers:
        return TypeConverters.serialize(arg)
    return quote_plus(str(arg))


def append_querystring(url: str, args: Dict[str, Any]):
    if args:
        for key in args:
            val = args[key]
            # print("append_querystring", key, val)
            if val is None:
                continue
            url += '&' if '?' in url else '?'
            qs_val = qsvalue(val)
            if qs_val is not None:
                url += key + '=' + qs_val
    return url


def has_request_body(method: str):
    return not (method == "GET" or method == "DELETE" or method == "HEAD" or method == "OPTIONS")


def combine_with(base_url: str, relative_url: str):
    if not base_url:
        return relative_url
    if not relative_url:
        return base_url
    # We want to handle relative urls different by maintaining the base_url path at all times
    temp_relative_url = relative_url
    temp_base_url = base_url
    if not temp_base_url.endswith("/"):
        temp_base_url += "/"
    if temp_relative_url.startswith("/"):
        temp_relative_url = temp_relative_url[1:]
    return urljoin(temp_base_url, temp_relative_url)

@dataclass_json
@dataclass
class SendContext:
    session: Optional[requests.Session] = None
    headers: Dict[str, str] = field(default_factory=dict)
    method: str = None
    url: Optional[str] = None
    request: Optional[Union[IReturn, IReturnVoid, List[IReturn], List[IReturnVoid]]] = None
    body: Optional[Any] = None
    body_string: Optional[str] = None
    args: Optional[Dict[str, str]] = None
    response_as: type = None
    request_filter: Callable[[Any], None] = None
    response_filter: Callable[[Response], None] = None

    def exec(self):
        if has_request_body(self.method):
            if CONTENT_TYPE not in self.headers:
                self.headers[CONTENT_TYPE] = JSON_MIME_TYPE
        else:
            if CONTENT_TYPE in self.headers:
                self.headers.pop(CONTENT_TYPE)

        if Log.debug_enabled():
            using = "requests"
            if self.session is not None:
                using = "session"
                Log.debug(f"{using}.cookies: {self.session.cookies}")
                # if "ss-tok" in self.session.cookies:
                #     ss_tok = self.session.cookies["ss-tok"]
                #     print('ss_tok', inspect_jwt(ss_tok))
            Log.debug(
                f"{using}.request({self.method}): url={self.url}, headers={self.headers}, data={self.body_string}")

        if has_request_body(self.method):
            if self.session is not None:
                response = self.session.request(self.method, self.url, data=self.body_string, headers=self.headers)
            else:
                response = requests.request(self.method, self.url, data=self.body_string, headers=self.headers)
        else:
            if self.session is not None:
                response = self.session.request(self.method, self.url, headers=self.headers)
            else:
                response = requests.request(self.method, self.url, headers=self.headers)

        if Log.debug_enabled():
            if self.response_as is not bytes:
                Log.debug(f"response.text: {response.text}")
            else:
                Log.debug(f"response.content len: {len(response.content)}")
            Log.debug(f"response.cookies[{len(response.cookies)}]: {response.cookies}")

        return response


class WebServiceExceptionType(Enum):
    DEFAULT = 1
    REFRESH_TOKEN_EXCEPTION = 2


class WebServiceException(Exception):
    status_code: int = None
    status_description: str = None
    message: str = None
    inner_exception: Exception = None
    response_status: ResponseStatus = None
    type: WebServiceExceptionType = WebServiceExceptionType.DEFAULT


T = TypeVar('T')

@dataclass
class UploadFile:
    field_name: Optional[str]
    file_name: Optional[str]
    content_type: Optional[str]
    stream: BinaryIO

    def __post_init__(self):
        if not self.content_type and self.file_name:
            guessed_type = mimetypes.guess_type(self.file_name)[0]
            self.content_type = guessed_type or 'application/octet-stream'

class JsonServiceClient:
    base_url: str = None
    reply_base_url: str = None
    oneway_base_url: str = None
    headers: Optional[Dict[str, str]] = None
    bearer_token: Optional[str] = None
    refresh_token: Optional[str] = None
    refresh_token_uri: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    use_token_cookie = False
    global_request_filter: Callable[[SendContext], None] = None  # static
    request_filter: Callable[[SendContext], None] = None
    global_response_filter: Callable[[Response], None] = None  # static
    response_filter: Callable[[Response], None] = None
    exception_filter: Callable[[Response, Exception], None] = None
    global_exception_filter: Callable[[Response, Exception], None] = None
    on_authentication_required: Callable[[], None] = None

    _session: requests.Session = None

    def __init__(self, base_url:str):
        if not base_url:
            raise TypeError(f"base_url is required")
        self.base_url = base_url
        self._session = requests.Session()
        self.headers = {'Accept': JSON_MIME_TYPE}
        self.set_base_path('api')

    def set_base_path(self, base_path:str=''):
        if not base_path:
            self.reply_base_url = combine_with(self.base_url, 'json/reply') + "/"
            self.oneway_base_url = combine_with(self.base_url, 'json/oneway') + "/"
        else:
            self.reply_base_url = combine_with(self.base_url, base_path) + "/"
            self.oneway_base_url = combine_with(self.base_url, base_path) + "/"
        return self

    def set_credentials(self, username:str, password:str):
        self.username = username
        self.password = password
        return self

    def set_bearer_token(self, bearer_token:str):
        self.bearer_token = bearer_token
        return self

    def set_refresh_token(self, refresh_token:str):
        self.refresh_token = refresh_token
        return self

    def _get_cookie_value(self, name: str) -> Optional[str]:
        if self._session is not None and name in self._session.cookies:
            return self._session.cookies[name]
        return None

    @property
    def token_cookie(self):
        return self._get_cookie_value(SS_TOKEN_COOKIE)

    @property
    def refresh_token_cookie(self):
        return self._get_cookie_value(SS_REFRESH_TOKEN_COOKIE)

    def create_url_from_dto(self, method: str, request: Any):
        url = combine_with(self.reply_base_url, nameof(request))
        if not has_request_body(method):
            url = append_querystring(url, to_dict(request, key_case=clean_camelcase))
        return url

    def get(self, request: IReturn[T], args: Dict[str, Any] = None) -> T:
        return self.send(request, "GET", None, args)

    def post(self, request: IReturn[T], body: Any = None, args: Dict[str, Any] = None) -> T:
        return self.send(request, "POST", body, args)

    def put(self, request: IReturn[T], body: Any = None, args: Dict[str, Any] = None) -> T:
        return self.send(request, "PUT", body, args)

    def patch(self, request: IReturn[T], body: Any = None, args: Dict[str, Any] = None) -> T:
        return self.send(request, "PATCH", body, args)

    def delete(self, request: IReturn[T], args: Dict[str, Any] = None) -> T:
        return self.send(request, "DELETE", None, args)

    def options(self, request: IReturn[T], args: Dict[str, Any] = None) -> T:
        return self.send(request, "OPTIONS", None, args)

    def head(self, request: IReturn[T], args: Dict[str, Any] = None) -> T:
        return self.send(request, "HEAD", None, args)

    def to_absolute_url(self, path_or_url: str):
        if path_or_url.startswith("http://") or path_or_url.startswith("https://"):
            return path_or_url
        return combine_with(self.base_url, path_or_url)

    def get_url(self, path: str, response_as: Type, args: Dict[str, Any] = None):
        return self.send_url(path, "GET", response_as, None, args)

    def delete_url(self, path: str, response_as: Type, args: Dict[str, Any] = None):
        return self.send_url(path, "DELETE", response_as, None, args)

    def options_url(self, path: str, response_as: Type, args: Dict[str, Any] = None):
        return self.send_url(path, "OPTIONS", response_as, None, args)

    def head_url(self, path: str, response_as: Type, args: Dict[str, Any] = None):
        return self.send_url(path, "HEAD", response_as, None, args)

    def post_url(self, path: str, body: Any = None, response_as: Type = None, args: Dict[str, Any] = None):
        return self.send_url(path, "POST", response_as, body, args)

    def put_url(self, path: str, body: Any = None, response_as: Type = None, args: Dict[str, Any] = None):
        return self.send_url(path, "PUT", response_as, body, args)

    def patch_url(self, path: str, body: Any = None, response_as: Type = None, args: Dict[str, Any] = None):
        return self.send_url(path, "PATCH", response_as, body, args)

    def send_url(self, path: str, method: str = None, response_as: Type = None, body: Any = None,
                 args: Dict[str, Any] = None):

        if body and not response_as:
            response_as = _resolve_response_type(body)

        info = SendContext(
            session=self._session,
            headers=self.headers.copy(),
            method=method or resolve_httpmethod(body),
            url=self.to_absolute_url(path),
            request=None,
            body=body,
            body_string=None,
            args=args,
            response_as=response_as)

        return self.send_request(info)

    def send(self, request, method: Any = None, body: Any = None, args: Dict[str, Any] = None):
        if not isinstance(request, IReturn) and not isinstance(request, IReturnVoid):
            raise TypeError(f"'{nameof(request)}' does not implement IReturn or IReturnVoid")

        response_as = _resolve_response_type(request)
        if response_as is None:
            raise TypeError(f"Could not resolve Response Type for '{nameof(request)}'")

        return self.send_request(SendContext(
            session=self._session,
            headers=self.headers.copy(),
            method=method or resolve_httpmethod(request),
            url=None,
            request=request,
            body=body,
            body_string=None,
            args=args,
            response_as=response_as))

    def post_file_with_request(self, request: IReturn[T], file: UploadFile) -> T:
        """
        Post file with a request DTO using multipart/form-data

        :param request: The request DTO
        :param files: Single UploadFile or List of UploadFile objects
        :return: Response DTO
        """
        return self.post_files_with_request_url(combine_with(self.reply_base_url, nameof(request)), request, [files])

    def post_file_with_request_url(self, request_uri: str, request: IReturn[T], file: UploadFile) -> T:
        """
        Post file with a request DTO using multipart/form-data

        :param request_uri: The request URI
        :param request: The request DTO
        :param files: Single UploadFile or List of UploadFile objects
        :return: Response DTO
        """
        return self.post_files_with_request_url(combine_with(self.reply_base_url, nameof(request)), request, [files])

    def post_files_with_request(self, request: IReturn[T], files: List[UploadFile]) -> T:
        """
        Post files with a request DTO using multipart/form-data

        :param request: The request DTO
        :param files: Single UploadFile or List of UploadFile objects
        :return: Response DTO
        """
        return self.post_files_with_request_url(combine_with(self.reply_base_url, nameof(request)), request, files)

    def post_files_with_request_url(self, request_uri: str, request: Any, files: List[UploadFile]) -> T:
        """
        Post files with a request DTO using multipart/form-data

        :param request_uri: The request URI
        :param request: The request DTO
        :param files: Single UploadFile or List of UploadFile objects
        :return: Response DTO
        """
        if isinstance(files, UploadFile):
            files = [files]

        # Convert request DTO to dict
        request_dict = to_dict(request, key_case=clean_camelcase)

        # Prepare the files dictionary for requests
        files_dict = {}
        for file in files:
            files_dict[file.field_name or 'upload'] = (
                file.file_name,
                file.stream,
                file.content_type or 'application/octet-stream'
            )

        # Prepare headers
        headers = self.headers.copy()
        if self.bearer_token:
            headers['Authorization'] = f'Bearer {self.bearer_token}'

        url = self.to_absolute_url(request_uri)

        try:
            # Send the multipart request
            response = self._session.post(
                url,
                data=request_dict,
                files=files_dict,
                headers=headers,
                verify=self._session.verify
            )

            # Handle errors
            response.raise_for_status()

            # Parse response
            response_type = _resolve_response_type(request)
            if response_type is None:
                return response.json()

            if response_type is str:
                return response.text

            if response_type is bytes:
                return response.content

            return from_json(response_type, response.text)

        finally:
            # Close all file streams
            for file in files:
                file.stream.close()


    @staticmethod
    def assert_valid_batch_request(request_dtos: list):
        if not isinstance(request_dtos, list):
            raise TypeError(f"'{nameof(request_dtos)}' is not a List")

        if len(request_dtos) == 0:
            return []

        request = request_dtos[0]
        if not isinstance(request, IReturn) and not isinstance(request, IReturnVoid):
            raise TypeError(f"'{nameof(request)}' does not implement IReturn or IReturnVoid")

        item_response_as = _resolve_response_type(request)
        if item_response_as is None:
            raise TypeError(f"Could not resolve Response Type for '{nameof(request)}'")
        return request, item_response_as

    def send_all(self, request_dtos: List[IReturn[T]]):
        request, item_response_as = self.assert_valid_batch_request(request_dtos)
        url = combine_with(self.reply_base_url, nameof(request) + "[]")

        return self.send_request(SendContext(
            session=self._session,
            headers=self.headers.copy(),
            method="POST",
            url=url,
            request=list(request_dtos),
            body=None,
            body_string=None,
            args=None,
            response_as=list.__class_getitem__(item_response_as)))  # requires 3.9

    def send_all_oneway(self, request_dtos: list):
        request, item_response_as = self.assert_valid_batch_request(request_dtos)
        url = combine_with(self.oneway_base_url, nameof(request) + "[]")

        self.send_request(SendContext(
            session=self._session,
            headers=self.headers.copy(),
            method="POST",
            url=url,
            request=list(request_dtos),
            body=None,
            body_string=None,
            args=None,
            response_as=list.__class_getitem__(item_response_as)))  # requires 3.9

    def _resend_request(self, info: SendContext):

        if self.bearer_token is not None:
            info.headers[AUTHORIZATION_HEADER] = f"Bearer {self.bearer_token}"
        elif self.username is not None:
            info.headers[AUTHORIZATION_HEADER] = \
                "Basic " + base64.b64encode(f"{self.username}:{self.password}".encode('ascii')).decode('ascii')

        response = info.exec()

        try:
            response.raise_for_status()
            res_dto = self._create_response(response, info)
            return res_dto, response
        except Exception as e:
            raise self._handle_error(response, e)

    def _create_response(self, response: Response, info: SendContext):

        if info.response_filter:
            info.response_filter(response)
        if self.response_filter:
            self.response_filter(response)
        if JsonServiceClient.global_response_filter:
            JsonServiceClient.global_response_filter(response)

        if len(response.cookies) > 0 and SS_REFRESH_TOKEN_COOKIE in response.cookies:
            self.use_token_cookie = True

        into = info.response_as

        if into is bytes:
            return response.content

        json_str = response.text

        if into is None:
            return json.loads(json_str)

        if into is str:
            return json_str

        try:
            res_dto = from_json(into, json_str)
        except Exception as e:
            Log.error(f"Failed to deserialize into {into}: {e}", e)
            raise e

        return res_dto

    def _raise_error(self, res: Response, e: Exception) -> Exception:
        if self.exception_filter:
            self.exception_filter(res, e)
        if JsonServiceClient.global_exception_filter:
            JsonServiceClient.global_exception_filter(res, e)
        return e

    def _handle_error(self, hold_res: Optional[Response], e: Exception, kind: Optional[WebServiceExceptionType] = None):
        if type(e) == WebServiceException:
            raise self._raise_error(hold_res, e)

        web_ex = WebServiceException()
        web_ex.inner_exception = e
        web_ex.status_code = 500
        web_ex.status_description = ex_message(e)
        if kind is not None:
            web_ex.type = kind

        res = hold_res
        if type(e) == HTTPError and e.response is not None:
            res = e.response

        if res is not None:
            if Log.debug_enabled():
                Log.debug(f"error.text: {res.text}")
            web_ex.status_code = res.status_code
            web_ex.status_description = res.reason

            web_ex.response_status = ResponseStatus(
                error_code=f"{res.status_code}",
                message=res.reason)

            try:
                error_response: EmptyResponse = from_json(EmptyResponse, res.text)
                if error_response is not None:
                    web_ex.response_status = error_response.response_status
            except Exception as ex:
                Log.error(f"Could not deserialize error response {res.text}", ex)

        raise self._raise_error(res, web_ex)

    def create_request(self, info: SendContext):
        try:
            url = info.url
            body = info.body or info.request
            if not url:
                body_not_request_dto = info.request and info.body
                if body_not_request_dto:
                    url = combine_with(self.reply_base_url, nameof(info.request))
                    url = append_querystring(url, to_dict(info.request, key_case=clean_camelcase))
                else:
                    url = self.create_url_from_dto(info.method, body)

            if not url:
                raise TypeError

            if info.args:
                url = append_querystring(url, info.args)
        except Exception as e:
            if Log.debug_enabled():
                Log.debug(f"send_request(): {ex_message(e)}")
            return self._handle_error(None, e)

        info.url = url
        if info.request_filter:
            info.request_filter(info)
        if self.request_filter:
            self.request_filter(info)
        if JsonServiceClient.global_request_filter:
            JsonServiceClient.global_request_filter(info)

        if has_request_body(info.method):
            if type(body) is str:
                info.body_string = body
            else:
                info.body_string = to_json(body)

        return info

    def send_request(self, info: SendContext):
        info = self.create_request(info)
        if Log.debug_enabled():
            Log.debug(f"info method: {info.method}, url: {info.url}, body_string: {info.body_string}")
        response: Optional[Response] = None
        try:
            res_dto, response = self._resend_request(info)

            if Log.debug_enabled():
                Log.debug(f"res_dto = {type(res_dto)}")

            return res_dto
        except Exception as e:
            if Log.debug_enabled():
                Log.debug(f"send_request() create_response: {ex_message(e)}")

            has_refresh_token_cookie = False
            if self.refresh_token is not None or self.use_token_cookie or has_refresh_token_cookie:
                Log.debug("attempting to refresh bearer_token with refresh_token")
                jwt_request = GetAccessToken(refresh_token=self.refresh_token)
                url = self.refresh_token_uri or self.create_url_from_dto("POST", jwt_request)

                try:
                    jwt_info = SendContext(
                        session=self._session,
                        headers=self.headers.copy(),
                        method="POST",
                        request=jwt_request,
                        url=url,
                        response_as=_resolve_response_type(jwt_request))
                    jwt_info = self.create_request(jwt_info)
                    jwt_res = jwt_info.exec()
                    jwt_res.raise_for_status()
                    jwt_response: GetAccessTokenResponse = self._create_response(jwt_res, jwt_info)
                    self.bearer_token = jwt_response.access_token
                    Log.debug("send_request() bearer_token refreshed")
                    if AUTHORIZATION_HEADER in info.headers:
                        info.headers.pop(AUTHORIZATION_HEADER)
                    res_dto, response = self._resend_request(info)
                    return res_dto
                except Exception as jwt_ex:
                    if Log.debug_enabled():
                        Log.debug(f"send_request() jwt_ex: {jwt_ex}")
                    return self._handle_error(response, jwt_ex, WebServiceExceptionType.REFRESH_TOKEN_EXCEPTION)

            if self.on_authentication_required is not None:
                self.on_authentication_required()
                res_dto, response = self._resend_request(info)
                return res_dto

            return self._handle_error(response, e)




import base64
import collections.abc as collections
import decimal
import inspect
import json
import math
import numbers
import os
import pathlib
import platform
import typing
import sys
from dataclasses import fields, is_dataclass
from datetime import datetime, timedelta
from enum import Enum, IntEnum, EnumMeta
from functools import reduce
from types import MappingProxyType
from typing_extensions import Type, get_origin, get_args
from typing import Callable, ForwardRef, Union, TypeVar
from typing import List, Dict, Any

import marshmallow.fields as mf

from servicestack.log import Log
from servicestack.utils import to_timespan, to_datetime, to_bytearray, from_bytearray, from_datetime, from_timespan
from .utils import snakecase, uppercase, camelcase


def is_optional(cls: Type): return get_origin(cls) is Union and type(None) in get_args(cls)


def is_list(cls: Type): return cls == typing.List or cls == list or get_origin(cls) == list


def is_dict(cls: Type): return cls == typing.Dict or cls == dict or get_origin(cls) == dict or cls == MappingProxyType


def nameof(instance):
    return type(instance).__name__


def generic_args(cls: Type):
    if not hasattr(cls, '__args__'):
        raise TypeError(f"{cls} is not a Generic Type")
    return cls.__args__


def generic_arg(cls: Type): return generic_args(cls)[0]


def _get_type_vars_map(cls: Type, type_map=None):
    if type_map is None:
        type_map = {}
    if hasattr(cls, '__orig_bases__'):
        for base_cls in cls.__orig_bases__:
            _get_type_vars_map(base_cls, type_map)
    generic_def = get_origin(cls)
    if generic_def is not None:
        generic_type_args = get_args(cls)
        i = 0
        for t in generic_def.__parameters__:
            k = t
            if type(k) == TypeVar:
                k = k.__name__
            type_map[k] = generic_type_args[i]
            i += 1
    return type_map


def _dict_with_string_keys(d: dict):
    to = {}
    for k, v in d.items():
        to[f"{k}"] = v
    return to


def has_type_vars(cls: Type):
    if cls is None:
        return None
    return isinstance(cls, TypeVar) or any(isinstance(x, TypeVar) for x in cls.__args__)


def _empty(x):
    return x is None or x == {} or x == []


def _str(x: Any):
    if type(x) == str:
        return x
    return f"{x}"


def identity(x: Any): return x


_JSON_TYPES = {str, bool, int, float}
_BUILT_IN_TYPES = {str, bool, int, float, decimal.Decimal, datetime, timedelta, bytes, bytearray, complex}


def is_builtin(t: Type):
    try:
        return t in _BUILT_IN_TYPES or issubclass(t, Enum)
    except Exception as e:  # throws if t is not hashable
        return False


def to_dict(obj: Any, key_case: Callable[[str], str] = identity, remove_empty: bool = True):
    t = type(obj)
    if obj is None or is_builtin(t):
        return obj
    if is_list(t):
        to = []
        for o in obj:
            use_val = to_dict(o, key_case=key_case, remove_empty=remove_empty)
            if not remove_empty or use_val is not None:
                to.append(use_val)
    elif is_dict(t):
        to = {}
        for k, v in obj.items():
            use_key = key_case(_str(k))
            use_val = to_dict(v, key_case=key_case, remove_empty=remove_empty)
            if not remove_empty or use_val is not None:
                to[use_key] = use_val
    elif hasattr(obj, 'to_dict'):  # dataclass
        d = obj.to_dict()
        to = {}
        for k, v in d.items():
            use_key = key_case(_str(k))
            use_val = to_dict(v, key_case=key_case, remove_empty=remove_empty)
            if not remove_empty or use_val is not None:
                to[use_key] = use_val
    elif hasattr(obj, '__dict__'):
        to = to_dict(vars(obj), key_case=key_case, remove_empty=remove_empty)
    else:
        to = obj
    if remove_empty:
        return clean_any(to)
    return to


def _clean_list(d: list):
    return [v for v in (clean_any(v) for v in d) if not _empty(v)]


def _clean_dict(d: dict):
    return {k: v for k, v in ((k, clean_any(v)) for k, v in d.items()) if not _empty(v)}


def clean_any(d):
    """recursively remove empty lists, empty dicts, or None elements from a dictionary"""
    if is_dict(d):
        return _clean_dict(d)
    elif is_list(d):
        return _clean_list(d)
    else:
        return d


def _json_encoder(obj: Any):
    t = type(obj)
    if is_dataclass(t) or is_dict(t):
        return to_dict(obj)
    if isinstance(obj, datetime):
        return obj.isoformat()
    if isinstance(obj, timedelta):
        return to_timespan(obj)
    if isinstance(obj, bytes):
        return base64.b64encode(obj).decode('ascii')
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    if t in _JSON_TYPES:
        return obj
    if is_builtin(t):
        return _str(obj)
    raise TypeError(f"Unsupported Type in JSON encoding: {t}")


def to_json(obj: Any, indent=None):
    return json.dumps(to_dict(obj), indent=indent, default=_json_encoder)


class TypeConverters:
    serializers = {}    # Py3.9: dict[Type, Callable[[Any], Any]]
    deserializers = {}  # Py3.9: dict[Type, Callable[[Any], Any]]

    @staticmethod
    def register(cls: Type, serializer: Callable[[Any], Any] = None, deserializer: Callable[[Any], Any] = None):
        if serializer is not None:
            TypeConverters.serializers[cls] = serializer
        if deserializer is not None:
            TypeConverters.deserializers[cls] = deserializer

    @staticmethod
    def serialize(obj: Any):
        cls = type(obj)
        if cls in TypeConverters.serializers:
            serializer = TypeConverters.serializers[cls]
            try:
                return serializer(obj)
            except Exception as e:
                Log.error(f"serializer(obj) {cls}({obj})", e)
                raise e

    @staticmethod
    def deserialize(cls: Type, obj: Any):
        deserializer = TypeConverters.deserializers[cls]
        try:
            return deserializer(obj)
        except Exception as e:
            Log.error(f"deserializer(obj) {cls}({obj})", e)
            raise e


TypeConverters.serializers = {
    datetime: to_datetime,
    timedelta: to_timespan,
    bytes: to_bytearray,
    bytearray: to_bytearray,
}
TypeConverters.deserializers = {
    mf.DateTime: from_datetime,
    mf.TimeDelta: from_timespan,
    datetime: from_datetime,
    timedelta: from_timespan,
    bytes: from_bytearray,
    bytearray: from_bytearray,
}


def _resolve_forwardref(cls: Type, orig: Type = None):
    type_name = cls.__forward_arg__
    if orig is not None and orig.__name__ == type_name:
        return orig
    if type_name not in globals():
        raise TypeError(f"Could not resolve ForwardRef('{type_name}')")
    return globals()[type_name]


def unwrap(cls: Type, module:str):
    if type(cls) == ForwardRef:
        cls = _resolve_forwardref(cls)
    if isinstance(cls, str):
        cls = eval(cls, {}, vars(sys.modules[module]))
    if is_optional(cls):
        return generic_arg(cls)
    return cls


def dict_get(name: str, obj: dict, case: Callable[[str], str] = None):
    if name in obj:
        return obj[name]
    if case:
        name_case = case(name)
        if name_case in obj:
            return obj[name_case]
    name_snake = snakecase(name)
    if name_snake in obj:
        return obj[name_snake]
    name_camel = camelcase(name)
    if name_camel in obj:
        return obj[name_camel]
    if name.endswith('_'):
        return dict_get(name.rstrip('_'), obj, case)
    return None


def sanitize_name(s: str):
    return s.replace('_', '').upper()


def enum_get(cls: Union[Enum, Type], key: Union[str, int]):
    if type(key) == int or issubclass(cls, IntEnum):
        return cls(key)
    try:
        return cls[key]
    except Exception as e:
        try:
            upper_snake = uppercase(snakecase(key))
            return cls[upper_snake]
        except Exception as e2:
            sanitize_key = sanitize_name(key)
            for value in cls.__members__.values():
                if sanitize_key == sanitize_name(value):
                    return value
            for member in cls.__members__.keys():
                if sanitize_key == sanitize_name(member):
                    return cls[member]
    raise TypeError(f"{key} is not a member of {nameof(Enum)}")


def _resolve_type(cls: Type, substitute_types: Dict[Type, type] ):
    if type(cls) == TypeVar:
        # TypeVar('T') == TypeVar('T') is false, 
        # I think this should work
        cls = cls.__name__
    if substitute_types is None:
        return cls
    return substitute_types[cls] if cls in substitute_types else cls


def convert(into: Type, obj: Any, substitute_types: Dict[Type, type] = None, module = None):
    if obj is None:
        return None
    into = unwrap(into, module)
    into = _resolve_type(into, substitute_types)
    if Log.debug_enabled():
        Log.debug(f"convert({into}, {substitute_types}, {obj})")

    is_type = type(into) == type
    if not is_type:
        Log.debug(f"type of {into} is not a class")

    generic_def = get_origin(into)
    if generic_def is not None and is_dataclass(generic_def):
        reified_types = _get_type_vars_map(into)
        return convert(generic_def, obj, reified_types)

    if is_dataclass(into):
        to = {}
        for f in fields(into):
            val = dict_get(f.name, obj)
            # print(f"to[{f.name}] = convert({f.type}, {val}, {substitute_types})")
            to[f.name] = convert(f.type, val, substitute_types, into.__module__)
            # print(f"to[{f.name}] = {to[f.name]}")
        return into(**to)
    elif is_list(into):
        el_type = _resolve_type(generic_arg(into), substitute_types)
        to = []
        for item in obj:
            to.append(convert(el_type, item, substitute_types, into.__module__))
        return to
    elif is_dict(into):
        key_type, val_type = generic_args(into)
        key_type = _resolve_type(key_type, substitute_types)
        val_type = _resolve_type(val_type, substitute_types)
        to = {}
        if not hasattr(obj, 'items'):
            Log.warn(f"dict {obj} ({type(type)}) does not have items()")
        for key, val in obj.items():
            to_key = convert(key_type, key, substitute_types, into.__module__)
            to_val = convert(val_type, val, substitute_types, into.__module__)
            to[to_key] = to_val
        return to
    else:
        if into in TypeConverters.deserializers:
            converter = TypeConverters.deserializers[into]
            try:
                return converter(obj)
            except Exception as e:
                Log.error(f"converter(obj) {into}({obj})", e)
                raise e
        elif inspect.isclass(into) and issubclass(into, mf.Field):
            try:
                return into().deserialize(obj)
            except Exception as e:
                Log.error(f"into().deserialize(obj) {into}({obj})", e)
                raise e
        elif is_type and (issubclass(into, Enum) or into == EnumMeta):
            try:
                return enum_get(into, obj)
            except Exception as e:
                print(into, type(into), obj, type(obj))
                Log.error(f"Enum into[obj] {into}[{obj}]", e)
                raise e
        else:
            try:
                return into(obj)
            except Exception as e:
                # if into == typing.Dict or get_origin(into) == typing.Dict:
                #     print("WAS A typing.Dict")
                Log.error(f"into(obj) {into}({obj})", e)
                raise e


def from_json(into: Type, json_str: str):
    if json_str is None or json_str == "":
        return None
    json_obj = json.loads(json_str)
    return convert(into, json_obj)


# inspect utils
def all_keys(obj):
    keys = []
    if not isinstance(obj, collections.Iterable):
        return keys
    for o in obj:
        if is_builtin(type(o)):
            continue
        for key in o:
            key = _str(key)
            if key is not None and key not in keys:
                keys.append(key)
    return keys


def inspect_vars(objs):
    if not isinstance(objs, dict):
        raise TypeError('objs must be a dictionary')

    to = to_dict(objs)

    inspect_vars_path = os.environ.get('INSPECT_VARS')
    if inspect_vars_path is None:
        return
    if platform.system() == 'Windows':
        inspect_vars_path = inspect_vars_path.replace("/", "\\")
    else:
        inspect_vars_path = inspect_vars_path.replace("\\", "/")

    pathlib.Path(os.path.dirname(inspect_vars_path)).mkdir(parents=True, exist_ok=True)

    with open(inspect_vars_path, 'w') as outfile:
        json_str = to_json(to)
        outfile.write(json_str)


def dump(obj):
    return to_json(obj, indent=4).replace('"', '').replace(': null', ':')


def printdump(obj):
    print(dump(obj))


def _align_left(s: str, length: int, pad: str = ' '):
    if length < 0:
        return ""
    alen = length + 1 - len(s)
    if alen <= 0:
        return s
    return pad + s + (pad * (length + 1 - len(s)))


def _align_center(s: str, length: int, pad: str = ' '):
    if length < 0:
        return ""
    nlen = len(s)
    half = math.floor(length / 2 - nlen / 2)
    odds = abs((nlen % 2) - (length % 2))
    return (pad * (half + 1)) + s + (pad * (half + 1 + odds))


def _align_right(s: str, length: int, pad: str = ' '):
    if length < 0:
        return ""
    alen = length + 1 - len(s)
    if alen <= 0:
        return s
    return (pad * (length + 1 - len(s))) + s + pad


def _align_auto(obj: Any, length: int, pad: str = ' '):
    s = _str(obj)
    if len(s) <= length:
        if isinstance(obj, numbers.Number):
            return _align_right(s, length, pad)
        return _align_left(s, length, pad)
    return s


def sanitize_key(key: str): return key.replace("_", " ").replace(" ", "").lower()


def lenient_getitem(obj: dict, key: str):
    if key in obj:
        return obj[key]
    sanitized_key = sanitize_key(key)
    for k in obj:
        if sanitize_key(k) == sanitized_key:
            return obj[k]
    return ""


def table(objs, headers=None):
    if not is_list(type(objs)):
        raise TypeError('objs must be a list')
    map_rows = to_dict(objs)
    if headers is None:
        headers = all_keys(map_rows)
    col_sizes: Dict[str, int] = {}

    for k in headers:
        max = len(k)
        for row in map_rows:
            col = lenient_getitem(row, k)
            if col != "":
                val_size = len(_str(col))
                if val_size > max:
                    max = val_size
            col_sizes[k] = max

    # sum + ' padding ' + |
    row_width = reduce(lambda x, y: x + y, col_sizes.values(), 0) + \
                (len(col_sizes) * 2) + \
                (len(col_sizes) + 1)
    sb: List[str] = [f"+{'-' * (row_width - 2)}+"]
    head = "|"
    for k in headers:
        head += _align_center(k, col_sizes[k]) + "|"
    sb.append(head)
    sb.append(f"|{'-' * (row_width - 2)}|")

    for row in map_rows:
        to = "|"
        for k in headers:
            to += '' + _align_auto(lenient_getitem(row, k), col_sizes[k]) + "|"
        sb.append(to)

    sb.append(f"+{'-' * (row_width - 2)}+")
    return '\n'.join(sb)


def printtable(obj, headers=None):
    print(table(obj, headers))


def htmllist(d: dict):
    sb: List[str] = ["<table><tbody>"]
    for k, v in d.items():
        sb.append(f"<tr><th>{_str(k)}</th><td>{htmldump(v)}</td></tr>")
    sb.append("</tbody></table>")
    return ''.join(sb)


def htmldump(objs, headers=None):
    if is_builtin(type(objs)):
        return _str(objs)

    map_rows = to_dict(objs)
    t = type(map_rows)
    if is_dict(t):
        return htmllist(map_rows)

    if headers is None:
        headers = all_keys(map_rows)

    # print(headers, map_rows)

    sb: List[str] = ["<table>"]
    row = ["<thead><tr>"]
    for k in headers:
        row.append(f"<th>{k}</th>")
    row.append("</tr></head>")
    if len(row) > 2:
        sb.append(''.join(row))
    sb.append("<tbody>")

    rows = []
    for item in map_rows:
        rows.append("<tr>")
        if len(headers) > 0:
            row = []
            for k in headers:
                val = lenient_getitem(item, k)
                row.append(f"<td>{htmldump(val)}</td>")
            rows.append(''.join(row))
        else:
            rows.append(f"<td>{htmldump(item)}</td>")
        rows.append("</tr>")

    sb.append(''.join(rows))
    sb.append("</tbody>")
    sb.append("</table>")
    return '\n'.join(sb)


def printhtmldump(obj, headers=None):
    print(htmldump(obj, headers))



import os

import urllib3

from servicestack import JsonServiceClient, Authenticate

# TEST_URL = "https://localhost:5001"
# TEST_URL = "http://localhost:5000"
TEST_URL = "https://test.servicestack.net"
TECHSTACKS_URL = "https://techstacks.io"
# Load AI Server URL from environment variable
AI_SERVER_URL = os.getenv('AI_SERVER_URL')


def create_test_client():
    return JsonServiceClient(TEST_URL)


def create_techstacks_client():
    return JsonServiceClient(TECHSTACKS_URL)


def clear_session(client: JsonServiceClient):
    client.post(Authenticate(provider="logout"))


def create_aiserver_client():
    """Create a client for AI Server tests using environment variables"""
    url = os.getenv('AI_SERVER_URL')
    api_key = os.getenv('AI_SERVER_API_KEY')

    if not url or not api_key:
        raise EnvironmentError(
            "AI_SERVER_URL and AI_SERVER_API_KEY environment variables are required. "
            "Please set these before running the AI server tests."
        )

    client = JsonServiceClient(url)
    client.set_bearer_token(api_key)

    return client

def log():
    pass


""" Options:
Date: 2024-10-23 07:43:30
Version: 8.41
Tip: To override a DTO option, remove "#" prefix before updating
BaseUrl: https://techstacks.io

#GlobalNamespace: 
#AddServiceStackTypes: True
#AddResponseStatus: False
#AddImplicitVersion: 
#AddDescriptionAsComments: True
#IncludeTypes: 
#ExcludeTypes: 
#DefaultImports: datetime,decimal,marshmallow.fields:*,servicestack:*,typing:*,dataclasses:dataclass/field,dataclasses_json:dataclass_json/LetterCase/Undefined/config,enum:Enum/IntEnum
#DataClass: 
#DataClassJson: 
"""

import datetime
import decimal
from marshmallow.fields import *
from servicestack import *
from typing import *
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, LetterCase, Undefined, config
from enum import Enum, IntEnum


class PostType(str, Enum):
    ANNOUNCEMENT = 'Announcement'
    POST = 'Post'
    SHOWCASE = 'Showcase'
    QUESTION = 'Question'
    REQUEST = 'Request'


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Post:
    id: int = 0
    organization_id: int = 0
    user_id: int = 0
    type: Optional[PostType] = None
    category_id: int = 0
    title: Optional[str] = None
    slug: Optional[str] = None
    url: Optional[str] = None
    image_url: Optional[str] = None
    # @StringLength(2147483647)
    content: Optional[str] = None

    # @StringLength(2147483647)
    content_html: Optional[str] = None

    pin_comment_id: Optional[int] = None
    technology_ids: Optional[List[int]] = None
    from_date: Optional[datetime.datetime] = None
    to_date: Optional[datetime.datetime] = None
    location: Optional[str] = None
    meta_type: Optional[str] = None
    meta: Optional[str] = None
    approved: bool = False
    up_votes: int = 0
    down_votes: int = 0
    points: int = 0
    views: int = 0
    favorites: int = 0
    subscribers: int = 0
    reply_count: int = 0
    comments_count: int = 0
    word_count: int = 0
    report_count: int = 0
    links_count: int = 0
    linked_to_count: int = 0
    score: int = 0
    rank: int = 0
    labels: Optional[List[str]] = None
    ref_user_ids: Optional[List[int]] = None
    ref_links: Optional[List[str]] = None
    mute_user_ids: Optional[List[int]] = None
    last_comment_date: Optional[datetime.datetime] = None
    last_comment_id: Optional[int] = None
    last_comment_user_id: Optional[int] = None
    deleted: Optional[datetime.datetime] = None
    deleted_by: Optional[str] = None
    locked: Optional[datetime.datetime] = None
    locked_by: Optional[str] = None
    hidden: Optional[datetime.datetime] = None
    hidden_by: Optional[str] = None
    status: Optional[str] = None
    status_date: Optional[datetime.datetime] = None
    status_by: Optional[str] = None
    archived: bool = False
    bumped: Optional[datetime.datetime] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    created_by: Optional[str] = None
    modified: datetime.datetime = datetime.datetime(1, 1, 1)
    modified_by: Optional[str] = None
    ref_id: Optional[int] = None
    ref_source: Optional[str] = None
    ref_urn: Optional[str] = None


class ReportAction(str, Enum):
    DISMISS = 'Dismiss'
    DELETE = 'Delete'


class FlagType(str, Enum):
    VIOLATION = 'Violation'
    SPAM = 'Spam'
    ABUSIVE = 'Abusive'
    CONFIDENTIAL = 'Confidential'
    OFF_TOPIC = 'OffTopic'
    OTHER = 'Other'


class Frequency(IntEnum):
    DAILY = 1
    WEEKLY = 7
    MONTHLY = 30
    QUARTERLY = 90


class TechnologyTier(str, Enum):
    PROGRAMMING_LANGUAGE = 'ProgrammingLanguage'
    CLIENT = 'Client'
    HTTP = 'Http'
    SERVER = 'Server'
    DATA = 'Data'
    SOFTWARE_INFRASTRUCTURE = 'SoftwareInfrastructure'
    OPERATING_SYSTEM = 'OperatingSystem'
    HARDWARE_INFRASTRUCTURE = 'HardwareInfrastructure'
    THIRD_PARTY_SERVICES = 'ThirdPartyServices'


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyBase:
    id: int = 0
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    vendor_url: Optional[str] = None
    product_url: Optional[str] = None
    logo_url: Optional[str] = None
    description: Optional[str] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    created_by: Optional[str] = None
    last_modified: datetime.datetime = datetime.datetime(1, 1, 1)
    last_modified_by: Optional[str] = None
    owner_id: Optional[str] = None
    slug: Optional[str] = None
    logo_approved: bool = False
    is_locked: bool = False
    tier: Optional[TechnologyTier] = None
    last_status_update: Optional[datetime.datetime] = None
    organization_id: Optional[int] = None
    comments_post_id: Optional[int] = None
    view_count: int = 0
    fav_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Technology(TechnologyBase):
    pass


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyView:
    id: Optional[int] = None
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    vendor_url: Optional[str] = None
    product_url: Optional[str] = None
    logo_url: Optional[str] = None
    description: Optional[str] = None
    created: Optional[datetime.datetime] = None
    created_by: Optional[str] = None
    last_modified: Optional[datetime.datetime] = None
    last_modified_by: Optional[str] = None
    owner_id: Optional[str] = None
    slug: Optional[str] = None
    logo_approved: Optional[bool] = None
    is_locked: Optional[bool] = None
    tier: Optional[TechnologyTier] = None
    last_status_update: Optional[datetime.datetime] = None
    organization_id: Optional[int] = None
    comments_post_id: Optional[int] = None
    view_count: Optional[int] = None
    fav_count: Optional[int] = None


class IRegisterStats:
    pass


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyStackBase:
    id: int = 0
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    description: Optional[str] = None
    app_url: Optional[str] = None
    screenshot_url: Optional[str] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    created_by: Optional[str] = None
    last_modified: datetime.datetime = datetime.datetime(1, 1, 1)
    last_modified_by: Optional[str] = None
    is_locked: bool = False
    owner_id: Optional[str] = None
    slug: Optional[str] = None
    # @StringLength(2147483647)
    details: Optional[str] = None

    # @StringLength(2147483647)
    details_html: Optional[str] = None

    last_status_update: Optional[datetime.datetime] = None
    organization_id: Optional[int] = None
    comments_post_id: Optional[int] = None
    view_count: int = 0
    fav_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyStack(TechnologyStackBase):
    pass


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyStackView:
    id: Optional[int] = None
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    description: Optional[str] = None
    app_url: Optional[str] = None
    screenshot_url: Optional[str] = None
    created: Optional[datetime.datetime] = None
    created_by: Optional[str] = None
    last_modified: Optional[datetime.datetime] = None
    last_modified_by: Optional[str] = None
    is_locked: Optional[bool] = None
    owner_id: Optional[str] = None
    slug: Optional[str] = None
    details: Optional[str] = None
    details_html: Optional[str] = None
    last_status_update: Optional[datetime.datetime] = None
    organization_id: Optional[int] = None
    comments_post_id: Optional[int] = None
    view_count: Optional[int] = None
    fav_count: Optional[int] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class PostComment:
    id: int = 0
    post_id: int = 0
    user_id: int = 0
    reply_id: Optional[int] = None
    # @StringLength(2147483647)
    content: Optional[str] = None

    # @StringLength(2147483647)
    content_html: Optional[str] = None

    score: int = 0
    rank: int = 0
    up_votes: int = 0
    down_votes: int = 0
    favorites: int = 0
    word_count: int = 0
    report_count: int = 0
    deleted: Optional[datetime.datetime] = None
    hidden: Optional[datetime.datetime] = None
    modified: datetime.datetime = datetime.datetime(1, 1, 1)
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    created_by: Optional[str] = None
    ref_id: Optional[int] = None
    ref_source: Optional[str] = None
    ref_urn: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Organization:
    id: int = 0
    name: Optional[str] = None
    slug: Optional[str] = None
    description: Optional[str] = None
    description_html: Optional[str] = None
    color: Optional[str] = None
    text_color: Optional[str] = None
    link_color: Optional[str] = None
    background_color: Optional[str] = None
    background_url: Optional[str] = None
    logo_url: Optional[str] = None
    hero_url: Optional[str] = None
    lang: Optional[str] = None
    default_post_type: Optional[str] = None
    default_subscription_post_types: Optional[List[str]] = None
    post_types: Optional[List[str]] = None
    moderator_post_types: Optional[List[str]] = None
    delete_posts_with_report_count: int = 0
    disable_invites: Optional[bool] = None
    up_votes: int = 0
    down_votes: int = 0
    views: int = 0
    favorites: int = 0
    subscribers: int = 0
    comments_count: int = 0
    posts_count: int = 0
    score: int = 0
    rank: int = 0
    ref_id: Optional[int] = None
    ref_source: Optional[str] = None
    hidden: Optional[datetime.datetime] = None
    hidden_by: Optional[str] = None
    locked: Optional[datetime.datetime] = None
    locked_by: Optional[str] = None
    deleted: Optional[datetime.datetime] = None
    deleted_by: Optional[str] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    created_by: Optional[str] = None
    modified: datetime.datetime = datetime.datetime(1, 1, 1)
    modified_by: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class OrganizationLabel:
    slug: Optional[str] = None
    organization_id: int = 0
    description: Optional[str] = None
    color: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Category:
    id: int = 0
    organization_id: int = 0
    name: Optional[str] = None
    slug: Optional[str] = None
    description: Optional[str] = None
    color: Optional[str] = None
    technology_ids: Optional[List[int]] = None
    comments_count: int = 0
    posts_count: int = 0
    score: int = 0
    rank: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class OrganizationMember:
    id: int = 0
    organization_id: int = 0
    user_id: int = 0
    user_name: Optional[str] = None
    is_owner: bool = False
    is_moderator: bool = False
    deny_all: bool = False
    deny_posts: bool = False
    deny_comments: bool = False
    notes: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class OrganizationMemberInvite:
    id: int = 0
    organization_id: int = 0
    user_id: int = 0
    user_name: Optional[str] = None
    dismissed: Optional[datetime.datetime] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class PostReportInfo:
    id: int = 0
    organization_id: int = 0
    post_id: int = 0
    user_id: int = 0
    user_name: Optional[str] = None
    flag_type: Optional[FlagType] = None
    report_notes: Optional[str] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    acknowledged: Optional[datetime.datetime] = None
    acknowledged_by: Optional[str] = None
    dismissed: Optional[datetime.datetime] = None
    dismissed_by: Optional[str] = None
    title: Optional[str] = None
    report_count: int = 0
    created_by: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class PostCommentReportInfo:
    id: int = 0
    organization_id: int = 0
    post_id: int = 0
    post_comment_id: int = 0
    user_id: int = 0
    user_name: Optional[str] = None
    flag_type: Optional[FlagType] = None
    report_notes: Optional[str] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    acknowledged: Optional[datetime.datetime] = None
    acknowledged_by: Optional[str] = None
    dismissed: Optional[datetime.datetime] = None
    dismissed_by: Optional[str] = None
    content_html: Optional[str] = None
    report_count: int = 0
    created_by: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserRef:
    id: int = 0
    user_name: Optional[str] = None
    email: Optional[str] = None
    ref_id: Optional[int] = None
    ref_source: Optional[str] = None
    ref_urn: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class OrganizationSubscription:
    id: int = 0
    organization_id: int = 0
    user_id: int = 0
    user_name: Optional[str] = None
    post_types: Optional[List[str]] = None
    frequency_days: Optional[int] = None
    last_synced_id: Optional[int] = None
    last_synced: Optional[datetime.datetime] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserActivity:
    id: int = 0
    user_name: Optional[str] = None
    karma: int = 0
    technology_count: int = 0
    tech_stacks_count: int = 0
    posts_count: int = 0
    post_up_votes: int = 0
    post_down_votes: int = 0
    comment_up_votes: int = 0
    comment_down_votes: int = 0
    post_comments_count: int = 0
    pinned_comment_count: int = 0
    post_report_count: int = 0
    post_comment_report_count: int = 0
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    modified: datetime.datetime = datetime.datetime(1, 1, 1)


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyHistory(TechnologyBase):
    technology_id: int = 0
    operation: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyStackHistory(TechnologyStackBase):
    technology_stack_id: int = 0
    operation: Optional[str] = None
    technology_ids: Optional[List[int]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserInfo:
    user_name: Optional[str] = None
    avatar_url: Optional[str] = None
    stacks_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyInfo:
    tier: Optional[TechnologyTier] = None
    slug: Optional[str] = None
    name: Optional[str] = None
    logo_url: Optional[str] = None
    stacks_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechnologyInStack(TechnologyBase):
    technology_id: int = 0
    technology_stack_id: int = 0
    justification: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class TechStackDetails(TechnologyStackBase):
    technology_choices: Optional[List[TechnologyInStack]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LabelInfo:
    slug: Optional[str] = None
    color: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CategoryInfo:
    id: int = 0
    name: Optional[str] = None
    slug: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class OrganizationInfo:
    id: int = 0
    name: Optional[str] = None
    slug: Optional[str] = None
    ref_id: Optional[int] = None
    ref_source: Optional[str] = None
    up_votes: Optional[int] = None
    down_votes: Optional[int] = None
    members_count: int = 0
    rank: int = 0
    disable_invites: Optional[bool] = None
    lang: Optional[str] = None
    post_types: Optional[List[str]] = None
    moderator_post_types: Optional[List[str]] = None
    locked: Optional[datetime.datetime] = None
    labels: Optional[List[LabelInfo]] = None
    categories: Optional[List[CategoryInfo]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Option:
    name: Optional[str] = None
    title: Optional[str] = None
    value: Optional[TechnologyTier] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class HelloResponse:
    result: Optional[str] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationResponse:
    cache: int = 0
    id: int = 0
    slug: Optional[str] = None
    organization: Optional[Organization] = None
    labels: Optional[List[OrganizationLabel]] = None
    categories: Optional[List[Category]] = None
    owners: Optional[List[OrganizationMember]] = None
    moderators: Optional[List[OrganizationMember]] = None
    members_count: int = 0
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationMembersResponse:
    organization_id: int = 0
    results: Optional[List[OrganizationMember]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationAdminResponse:
    labels: Optional[List[OrganizationLabel]] = None
    members: Optional[List[OrganizationMember]] = None
    member_invites: Optional[List[OrganizationMemberInvite]] = None
    reported_posts: Optional[List[PostReportInfo]] = None
    reported_post_comments: Optional[List[PostCommentReportInfo]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateOrganizationForTechnologyResponse:
    organization_id: int = 0
    organization_slug: Optional[str] = None
    comments_post_id: int = 0
    comments_post_slug: Optional[str] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateOrganizationResponse:
    id: int = 0
    slug: Optional[str] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class OrganizationLabelResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AddOrganizationCategoryResponse:
    id: int = 0
    slug: Optional[str] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationCategoryResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AddOrganizationMemberResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationMemberResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class SetOrganizationMembersResponse:
    user_ids_added: Optional[List[int]] = None
    user_ids_removed: Optional[List[int]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationMemberInvitesResponse:
    results: Optional[List[OrganizationMemberInvite]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RequestOrganizationMemberInviteResponse:
    organization_id: int = 0
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationMemberInviteResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetPostResponse:
    cache: int = 0
    post: Optional[Post] = None
    comments: Optional[List[PostComment]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreatePostResponse:
    id: int = 0
    slug: Optional[str] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdatePostResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeletePostResponse:
    id: int = 0
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreatePostCommentResponse:
    id: int = 0
    post_id: int = 0
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdatePostCommentResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeletePostCommentResponse:
    id: int = 0
    post_id: int = 0
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserPostCommentVotesResponse:
    post_id: int = 0
    up_voted_comment_ids: Optional[List[int]] = None
    down_voted_comment_ids: Optional[List[int]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class PinPostCommentResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUsersByEmailsResponse:
    results: Optional[List[UserRef]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserPostActivityResponse:
    up_voted_post_ids: Optional[List[int]] = None
    down_voted_post_ids: Optional[List[int]] = None
    favorite_post_ids: Optional[List[int]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserOrganizationsResponse:
    members: Optional[List[OrganizationMember]] = None
    member_invites: Optional[List[OrganizationMemberInvite]] = None
    subscriptions: Optional[List[OrganizationSubscription]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostVoteResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostFavoriteResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostReportResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostCommentVoteResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostCommentReportResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class SessionInfoResponse:
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    id: Optional[str] = None
    referrer_url: Optional[str] = None
    user_auth_id: Optional[str] = None
    user_auth_name: Optional[str] = None
    user_name: Optional[str] = None
    display_name: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    email: Optional[str] = None
    created_at: datetime.datetime = datetime.datetime(1, 1, 1)
    last_modified: datetime.datetime = datetime.datetime(1, 1, 1)
    roles: Optional[List[str]] = None
    permissions: Optional[List[str]] = None
    is_authenticated: bool = False
    auth_provider: Optional[str] = None
    profile_url: Optional[str] = None
    github_profile_url: Optional[str] = None
    twitter_profile_url: Optional[str] = None
    access_token: Optional[str] = None
    avatar_url: Optional[str] = None
    tech_stacks: Optional[List[TechnologyStack]] = None
    favorite_tech_stacks: Optional[List[TechnologyStack]] = None
    favorite_technologies: Optional[List[Technology]] = None
    user_activity: Optional[UserActivity] = None
    members: Optional[List[OrganizationMember]] = None
    member_invites: Optional[List[OrganizationMemberInvite]] = None
    subscriptions: Optional[List[OrganizationSubscription]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyPreviousVersionsResponse:
    results: Optional[List[TechnologyHistory]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetAllTechnologiesResponse:
    results: Optional[List[Technology]] = None
    total: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyResponse:
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    technology: Optional[Technology] = None
    technology_stacks: Optional[List[TechnologyStack]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyFavoriteDetailsResponse:
    users: Optional[List[str]] = None
    favorite_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateTechnologyResponse:
    result: Optional[Technology] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateTechnologyResponse:
    result: Optional[Technology] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeleteTechnologyResponse:
    result: Optional[Technology] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyStackPreviousVersionsResponse:
    results: Optional[List[TechnologyStackHistory]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetPageStatsResponse:
    type: Optional[str] = None
    slug: Optional[str] = None
    view_count: int = 0
    fav_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class OverviewResponse:
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    top_users: Optional[List[UserInfo]] = None
    top_technologies: Optional[List[TechnologyInfo]] = None
    latest_tech_stacks: Optional[List[TechStackDetails]] = None
    popular_tech_stacks: Optional[List[TechnologyStack]] = None
    all_organizations: Optional[List[OrganizationInfo]] = None
    top_technologies_by_tier: Optional[Dict[str, List[TechnologyInfo]]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AppOverviewResponse:
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    all_tiers: Optional[List[Option]] = None
    top_technologies: Optional[List[TechnologyInfo]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetAllTechnologyStacksResponse:
    results: Optional[List[TechnologyStack]] = None
    total: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyStackResponse:
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    result: Optional[TechStackDetails] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyStackFavoriteDetailsResponse:
    users: Optional[List[str]] = None
    favorite_count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetConfigResponse:
    all_tiers: Optional[List[Option]] = None
    all_post_types: Optional[List[Option]] = None
    all_flag_types: Optional[List[Option]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateTechnologyStackResponse:
    result: Optional[TechStackDetails] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateTechnologyStackResponse:
    result: Optional[TechStackDetails] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeleteTechnologyStackResponse:
    result: Optional[TechStackDetails] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetFavoriteTechStackResponse:
    results: Optional[List[TechnologyStack]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class FavoriteTechStackResponse:
    result: Optional[TechnologyStack] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetFavoriteTechnologiesResponse:
    results: Optional[List[Technology]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class FavoriteTechnologyResponse:
    result: Optional[Technology] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserFeedResponse:
    results: Optional[List[TechStackDetails]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUsersKarmaResponse:
    results: Optional[Dict[int, int]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserInfoResponse:
    id: int = 0
    user_name: Optional[str] = None
    created: datetime.datetime = datetime.datetime(1, 1, 1)
    avatar_url: Optional[str] = None
    tech_stacks: Optional[List[TechnologyStack]] = None
    favorite_tech_stacks: Optional[List[TechnologyStack]] = None
    favorite_technologies: Optional[List[Technology]] = None
    user_activity: Optional[UserActivity] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LogoUrlApprovalResponse:
    result: Optional[Technology] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LockStackResponse:
    pass


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class EmailTestResponse:
    response_status: Optional[ResponseStatus] = None


# @Route("/ping")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Ping(IGet):
    pass


# @Route("/hello/{Name}")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Hello(IReturn[HelloResponse], IGet):
    name: Optional[str] = None


# @Route("/orgs/{Id}", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganization(IReturn[GetOrganizationResponse], IGet):
    id: Optional[int] = None


# @Route("/organizations/{Slug}", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationBySlug(IReturn[GetOrganizationResponse], IGet):
    slug: Optional[str] = None


# @Route("/orgs/{Id}/members", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationMembers(IReturn[GetOrganizationMembersResponse], IGet):
    id: int = 0


# @Route("/orgs/{Id}/admin", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationAdmin(IReturn[GetOrganizationAdminResponse], IGet):
    id: int = 0


# @Route("/orgs/posts/new", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateOrganizationForTechnology(IReturn[CreateOrganizationForTechnologyResponse], IPost):
    technology_id: Optional[int] = None
    tech_stack_id: Optional[int] = None


# @Route("/orgs", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateOrganization(IReturn[CreateOrganizationResponse], IPost):
    name: Optional[str] = None
    slug: Optional[str] = None
    description: Optional[str] = None
    ref_id: Optional[int] = None
    ref_source: Optional[str] = None
    ref_urn: Optional[str] = None


# @Route("/orgs/{Id}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganization(IReturn[UpdateOrganizationResponse], IPut):
    id: int = 0
    slug: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    color: Optional[str] = None
    text_color: Optional[str] = None
    link_color: Optional[str] = None
    background_color: Optional[str] = None
    background_url: Optional[str] = None
    logo_url: Optional[str] = None
    hero_url: Optional[str] = None
    lang: Optional[str] = None
    delete_posts_with_report_count: int = 0
    disable_invites: Optional[bool] = None
    default_post_type: Optional[str] = None
    default_subscription_post_types: Optional[List[str]] = None
    post_types: Optional[List[str]] = None
    moderator_post_types: Optional[List[str]] = None
    technology_ids: Optional[List[int]] = None


# @Route("/orgs/{Id}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeleteOrganization(IReturnVoid, IDelete):
    id: int = 0


# @Route("/orgs/{Id}/lock", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LockOrganization(IReturnVoid, IPut):
    id: int = 0
    lock: bool = False
    reason: Optional[str] = None


# @Route("/orgs/{OrganizationId}/labels", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AddOrganizationLabel(IReturn[OrganizationLabelResponse], IPost):
    organization_id: int = 0
    slug: Optional[str] = None
    description: Optional[str] = None
    color: Optional[str] = None


# @Route("/orgs/{OrganizationId}/members/{Slug}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationLabel(IReturn[OrganizationLabelResponse], IPut):
    organization_id: int = 0
    slug: Optional[str] = None
    description: Optional[str] = None
    color: Optional[str] = None


# @Route("/orgs/{OrganizationId}/labels/{Slug}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RemoveOrganizationLabel(IReturnVoid, IDelete):
    organization_id: int = 0
    slug: Optional[str] = None


# @Route("/orgs/{OrganizationId}/categories", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AddOrganizationCategory(IReturn[AddOrganizationCategoryResponse], IPost):
    organization_id: int = 0
    slug: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    technology_ids: Optional[List[int]] = None


# @Route("/orgs/{OrganizationId}/categories/{Id}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationCategory(IReturn[UpdateOrganizationCategoryResponse], IPut):
    organization_id: int = 0
    id: int = 0
    name: Optional[str] = None
    slug: Optional[str] = None
    description: Optional[str] = None
    technology_ids: Optional[List[int]] = None


# @Route("/orgs/{OrganizationId}/categories/{Id}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeleteOrganizationCategory(IReturnVoid, IDelete):
    organization_id: int = 0
    id: int = 0


# @Route("/orgs/{OrganizationId}/members", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AddOrganizationMember(IReturn[AddOrganizationMemberResponse], IPost):
    organization_id: int = 0
    user_name: Optional[str] = None
    is_owner: bool = False
    is_moderator: bool = False
    deny_posts: bool = False
    deny_comments: bool = False
    deny_all: bool = False
    notes: Optional[str] = None


# @Route("/orgs/{OrganizationId}/members/{Id}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationMember(IReturn[UpdateOrganizationMemberResponse], IPut):
    organization_id: int = 0
    user_id: int = 0
    is_owner: bool = False
    is_moderator: bool = False
    deny_posts: bool = False
    deny_comments: bool = False
    deny_all: bool = False
    notes: Optional[str] = None


# @Route("/orgs/{OrganizationId}/members/{UserId}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RemoveOrganizationMember(IReturnVoid, IDelete):
    organization_id: int = 0
    user_id: int = 0


# @Route("/orgs/{OrganizationId}/members/set", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class SetOrganizationMembers(IReturn[SetOrganizationMembersResponse], IPost):
    organization_id: int = 0
    github_user_names: Optional[List[str]] = None
    twitter_user_names: Optional[List[str]] = None
    emails: Optional[List[str]] = None
    remove_unspecified_members: bool = False
    is_owner: bool = False
    is_moderator: bool = False
    deny_posts: bool = False
    deny_comments: bool = False
    deny_all: bool = False


# @Route("/orgs/{OrganizationId}/invites", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetOrganizationMemberInvites(IReturn[GetOrganizationMemberInvitesResponse], IGet):
    organization_id: int = 0


# @Route("/orgs/{OrganizationId}/invites", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RequestOrganizationMemberInvite(IReturn[RequestOrganizationMemberInviteResponse], IPost):
    organization_id: int = 0


# @Route("/orgs/{OrganizationId}/invites/{UserId}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateOrganizationMemberInvite(IReturn[UpdateOrganizationMemberInviteResponse], IPut):
    organization_id: int = 0
    user_name: Optional[str] = None
    approve: bool = False
    dismiss: bool = False


# @Route("/posts", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryPosts(QueryDb[Post], IReturn[QueryResponse[Post]]):
    ids: Optional[List[int]] = None
    organization_id: Optional[int] = None
    organization_ids: Optional[List[int]] = None
    types: Optional[List[str]] = None
    any_technology_ids: Optional[List[int]] = None
    is_: Optional[List[str]] = field(metadata=config(field_name='is'), default=None)


# @Route("/posts/{Id}", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetPost(IReturn[GetPostResponse], IGet):
    id: int = 0
    include: Optional[str] = None


# @Route("/posts", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreatePost(IReturn[CreatePostResponse], IPost):
    organization_id: int = 0
    type: Optional[PostType] = None
    category_id: int = 0
    title: Optional[str] = None
    url: Optional[str] = None
    image_url: Optional[str] = None
    content: Optional[str] = None
    lock: Optional[bool] = None
    technology_ids: Optional[List[int]] = None
    labels: Optional[List[str]] = None
    from_date: Optional[datetime.datetime] = None
    to_date: Optional[datetime.datetime] = None
    meta_type: Optional[str] = None
    meta: Optional[str] = None
    ref_id: Optional[int] = None
    ref_source: Optional[str] = None
    ref_urn: Optional[str] = None


# @Route("/posts/{Id}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdatePost(IReturn[UpdatePostResponse], IPut):
    id: int = 0
    organization_id: int = 0
    type: Optional[PostType] = None
    category_id: int = 0
    title: Optional[str] = None
    url: Optional[str] = None
    image_url: Optional[str] = None
    content: Optional[str] = None
    lock: Optional[bool] = None
    technology_ids: Optional[List[int]] = None
    labels: Optional[List[str]] = None
    from_date: Optional[datetime.datetime] = None
    to_date: Optional[datetime.datetime] = None
    meta_type: Optional[str] = None
    meta: Optional[str] = None


# @Route("/posts/{Id}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeletePost(IReturn[DeletePostResponse], IDelete):
    id: int = 0


# @Route("/posts/{Id}/lock", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LockPost(IReturnVoid, IPut):
    id: int = 0
    lock: bool = False
    reason: Optional[str] = None


# @Route("/posts/{Id}/hide", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class HidePost(IReturnVoid, IPut):
    id: int = 0
    hide: bool = False
    reason: Optional[str] = None


# @Route("/posts/{Id}/status/{Status}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ChangeStatusPost(IReturnVoid, IPut):
    id: int = 0
    status: Optional[str] = None
    reason: Optional[str] = None


# @Route("/posts/{PostId}/report/{Id}", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ActionPostReport(IReturnVoid, IPost):
    post_id: int = 0
    id: int = 0
    report_action: Optional[ReportAction] = None


# @Route("/posts/{PostId}/comments", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreatePostComment(IReturn[CreatePostCommentResponse], IPost):
    post_id: int = 0
    reply_id: Optional[int] = None
    content: Optional[str] = None


# @Route("/posts/{PostId}/comments/{Id}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdatePostComment(IReturn[UpdatePostCommentResponse], IPut):
    id: int = 0
    post_id: int = 0
    content: Optional[str] = None


# @Route("/posts/{PostId}/comments/{Id}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeletePostComment(IReturn[DeletePostCommentResponse], IDelete):
    id: int = 0
    post_id: int = 0


# @Route("/posts/{PostId}/comments/{PostCommentId}/report/{Id}", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ActionPostCommentReport(IReturnVoid, IPost):
    id: int = 0
    post_comment_id: int = 0
    post_id: int = 0
    report_action: Optional[ReportAction] = None


# @Route("/user/comments/votes")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserPostCommentVotes(IReturn[GetUserPostCommentVotesResponse], IGet):
    post_id: int = 0


# @Route("/posts/{PostId}/comments/{Id}/pin", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class PinPostComment(IReturn[PinPostCommentResponse], IPut):
    id: int = 0
    post_id: int = 0
    pin: bool = False


# @Route("/users/by-email")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUsersByEmails(IReturn[GetUsersByEmailsResponse], IGet):
    emails: Optional[List[str]] = None


# @Route("/user/posts/activity")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserPostActivity(IReturn[GetUserPostActivityResponse], IGet):
    pass


# @Route("/user/organizations")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserOrganizations(IReturn[GetUserOrganizationsResponse], IGet):
    pass


# @Route("/posts/{Id}/vote", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostVote(IReturn[UserPostVoteResponse], IPut):
    id: int = 0
    weight: int = 0


# @Route("/posts/{Id}/favorite", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostFavorite(IReturn[UserPostFavoriteResponse], IPut):
    id: int = 0


# @Route("/posts/{Id}/report", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostReport(IReturn[UserPostReportResponse], IPut):
    id: int = 0
    flag_type: Optional[FlagType] = None
    report_notes: Optional[str] = None


# @Route("/posts/{PostId}/comments/{Id}", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostCommentVote(IReturn[UserPostCommentVoteResponse], IGet):
    id: int = 0
    post_id: int = 0
    weight: int = 0


# @Route("/posts/{PostId}/comments/{Id}/report", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserPostCommentReport(IReturn[UserPostCommentReportResponse], IPut):
    id: int = 0
    post_id: int = 0
    flag_type: Optional[FlagType] = None
    report_notes: Optional[str] = None


# @Route("/prerender/{**Path}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class StorePreRender(IReturnVoid, IPut):
    path: Optional[str] = None
    request_stream: Optional[bytes] = None


# @Route("/prerender/{**Path}", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetPreRender(IReturn[str], IGet):
    path: Optional[str] = None


# @Route("/my-session")
# @ValidateRequest(Validator="IsAuthenticated")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class SessionInfo(IReturn[SessionInfoResponse], IGet):
    pass


# @Route("/orgs/{OrganizationId}/subscribe", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class SubscribeToOrganization(IReturnVoid, IPut):
    organization_id: int = 0
    post_types: Optional[List[PostType]] = None
    frequency: Optional[Frequency] = None


# @Route("/posts/{PostId}/subscribe", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class SubscribeToPost(IReturnVoid, IPut):
    post_id: int = 0


# @Route("/orgs/{OrganizationId}/subscribe", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeleteOrganizationSubscription(IReturnVoid, IDelete):
    organization_id: int = 0


# @Route("/posts/{PostId}/subscribe", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeletePostSubscription(IReturnVoid, IDelete):
    post_id: int = 0


# @Route("/technology/{Slug}/previous-versions", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyPreviousVersions(IReturn[GetTechnologyPreviousVersionsResponse], IGet):
    slug: Optional[str] = None


# @Route("/technology", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetAllTechnologies(IReturn[GetAllTechnologiesResponse], IGet):
    pass


# @Route("/technology/search")
# @AutoQueryViewer(DefaultSearchField="Tier", DefaultSearchText="Data", DefaultSearchType="=", Description="Explore different Technologies", IconUrl="octicon:database", Title="Find Technologies")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class FindTechnologies(QueryDb2[Technology, TechnologyView], IReturn[QueryResponse[TechnologyView]]):
    ids: Optional[List[int]] = None
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    name_contains: Optional[str] = None
    vendor_name_contains: Optional[str] = None
    description_contains: Optional[str] = None


# @Route("/technology/query")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryTechnology(QueryDb2[Technology, TechnologyView], IReturn[QueryResponse[TechnologyView]]):
    ids: Optional[List[int]] = None
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    name_contains: Optional[str] = None
    vendor_name_contains: Optional[str] = None
    description_contains: Optional[str] = None


# @Route("/technology/{Slug}")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnology(IReturn[GetTechnologyResponse], IRegisterStats, IGet):
    slug: Optional[str] = None


# @Route("/technology/{Slug}/favorites")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyFavoriteDetails(IReturn[GetTechnologyFavoriteDetailsResponse], IGet):
    slug: Optional[str] = None


# @Route("/technology", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateTechnology(IReturn[CreateTechnologyResponse], IPost):
    name: Optional[str] = None
    slug: Optional[str] = None
    vendor_name: Optional[str] = None
    vendor_url: Optional[str] = None
    product_url: Optional[str] = None
    logo_url: Optional[str] = None
    description: Optional[str] = None
    is_locked: bool = False
    tier: Optional[TechnologyTier] = None


# @Route("/technology/{Id}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateTechnology(IReturn[UpdateTechnologyResponse], IPut):
    id: int = 0
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    vendor_url: Optional[str] = None
    product_url: Optional[str] = None
    logo_url: Optional[str] = None
    description: Optional[str] = None
    is_locked: bool = False
    tier: Optional[TechnologyTier] = None


# @Route("/technology/{Id}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeleteTechnology(IReturn[DeleteTechnologyResponse], IDelete):
    id: int = 0


# @Route("/techstacks/{Slug}/previous-versions", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyStackPreviousVersions(IReturn[GetTechnologyStackPreviousVersionsResponse], IGet):
    slug: Optional[str] = None


# @Route("/pagestats/{Type}/{Slug}")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetPageStats(IReturn[GetPageStatsResponse], IGet):
    type: Optional[str] = None
    slug: Optional[str] = None
    id: Optional[int] = None


# @Route("/techstacks/search")
# @AutoQueryViewer(DefaultSearchField="Description", DefaultSearchText="ServiceStack", DefaultSearchType="Contains", Description="Explore different Technology Stacks", IconUrl="material-icons:cloud", Title="Find Technology Stacks")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class FindTechStacks(QueryDb2[TechnologyStack, TechnologyStackView], IReturn[QueryResponse[TechnologyStackView]]):
    ids: Optional[List[int]] = None
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    name_contains: Optional[str] = None
    vendor_name_contains: Optional[str] = None
    description_contains: Optional[str] = None


# @Route("/techstacks/query")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryTechStacks(QueryDb2[TechnologyStack, TechnologyStackView], IReturn[QueryResponse[TechnologyStackView]]):
    ids: Optional[List[int]] = None
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    name_contains: Optional[str] = None
    vendor_name_contains: Optional[str] = None
    description_contains: Optional[str] = None


# @Route("/overview")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Overview(IReturn[OverviewResponse], IGet):
    reload: bool = False


# @Route("/app-overview")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AppOverview(IReturn[AppOverviewResponse], IGet):
    reload: bool = False


# @Route("/techstacks", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetAllTechnologyStacks(IReturn[GetAllTechnologyStacksResponse], IGet):
    pass


# @Route("/techstacks/{Slug}", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyStack(IReturn[GetTechnologyStackResponse], IRegisterStats, IGet):
    slug: Optional[str] = None


# @Route("/techstacks/{Slug}/favorites")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetTechnologyStackFavoriteDetails(IReturn[GetTechnologyStackFavoriteDetailsResponse], IGet):
    slug: Optional[str] = None


# @Route("/config")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetConfig(IReturn[GetConfigResponse], IGet):
    pass


# @Route("/techstacks", "POST")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CreateTechnologyStack(IReturn[CreateTechnologyStackResponse], IPost):
    name: Optional[str] = None
    slug: Optional[str] = None
    vendor_name: Optional[str] = None
    app_url: Optional[str] = None
    screenshot_url: Optional[str] = None
    description: Optional[str] = None
    details: Optional[str] = None
    is_locked: bool = False
    technology_ids: Optional[List[int]] = None


# @Route("/techstacks/{Id}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateTechnologyStack(IReturn[UpdateTechnologyStackResponse], IPut):
    id: int = 0
    name: Optional[str] = None
    vendor_name: Optional[str] = None
    app_url: Optional[str] = None
    screenshot_url: Optional[str] = None
    description: Optional[str] = None
    details: Optional[str] = None
    is_locked: bool = False
    technology_ids: Optional[List[int]] = None


# @Route("/techstacks/{Id}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DeleteTechnologyStack(IReturn[DeleteTechnologyStackResponse], IDelete):
    id: int = 0


# @Route("/favorites/techtacks", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetFavoriteTechStack(IReturn[GetFavoriteTechStackResponse], IGet):
    technology_stack_id: int = 0


# @Route("/favorites/techtacks/{TechnologyStackId}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AddFavoriteTechStack(IReturn[FavoriteTechStackResponse], IPut):
    technology_stack_id: int = 0


# @Route("/favorites/techtacks/{TechnologyStackId}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RemoveFavoriteTechStack(IReturn[FavoriteTechStackResponse], IDelete):
    technology_stack_id: int = 0


# @Route("/favorites/technology", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetFavoriteTechnologies(IReturn[GetFavoriteTechnologiesResponse], IGet):
    technology_id: int = 0


# @Route("/favorites/technology/{TechnologyId}", "PUT")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AddFavoriteTechnology(IReturn[FavoriteTechnologyResponse], IPut):
    technology_id: int = 0


# @Route("/favorites/technology/{TechnologyId}", "DELETE")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RemoveFavoriteTechnology(IReturn[FavoriteTechnologyResponse], IDelete):
    technology_id: int = 0


# @Route("/my-feed")
# @ValidateRequest(Validator="IsAuthenticated")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserFeed(IReturn[GetUserFeedResponse], IGet):
    pass


# @Route("/users/karma", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUsersKarma(IReturn[GetUsersKarmaResponse], IGet):
    user_ids: Optional[List[int]] = None


# @Route("/userinfo/{Id}")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetUserInfo(IReturn[GetUserInfoResponse], IGet):
    id: int = 0
    user_name: Optional[str] = None


# @Route("/users/{UserId}/avatar", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserAvatar(IGet):
    user_id: int = 0


# @Route("/mq/start")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class MqStart(IReturn[str]):
    pass


# @Route("/mq/stop")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class MqStop(IReturn[str]):
    pass


# @Route("/mq/stats")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class MqStats(IReturn[str]):
    pass


# @Route("/mq/status")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class MqStatus(IReturn[str]):
    pass


# @Route("/admin/technology/{TechnologyId}/logo")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LogoUrlApproval(IReturn[LogoUrlApprovalResponse], IPut):
    technology_id: int = 0
    approved: bool = False


# @Route("/admin/techstacks/{TechnologyStackId}/lock")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LockTechStack(IReturn[LockStackResponse], IPut):
    """
    Limit updates to TechStack to Owner or Admin users
    """

    # @Validate(Validator="GreaterThan(0)")
    technology_stack_id: int = 0

    is_locked: bool = False


# @Route("/admin/technology/{TechnologyId}/lock")
# @Api(Description="Limit updates to Technology to Owner or Admin users")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class LockTech(IReturn[LockStackResponse], IPut):
    """
    Limit updates to Technology to Owner or Admin users
    """

    # @Validate(Validator="GreaterThan(0)")
    technology_id: int = 0

    is_locked: bool = False


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class DummyTypes:
    post: Optional[List[Post]] = None


# @Route("/email/post/{PostId}")
# @ValidateRequest(Validator="IsAdmin")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class EmailTest(IReturn[EmailTestResponse]):
    post_id: Optional[int] = None


# @Route("/posts/comment", "GET")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryPostComments(QueryDb[PostComment], IReturn[QueryResponse[PostComment]]):
    id: Optional[int] = None
    user_id: Optional[int] = None
    post_id: Optional[int] = None
    content_contains: Optional[str] = None
    up_votes_above: Optional[int] = None
    up_votes_below: Optional[int] = None
    down_votes_above: Optional[int] = None
    down_votes: Optional[int] = None
    favorites_above: Optional[int] = None
    favorites_below: Optional[int] = None
    word_count_above: Optional[int] = None
    word_count_below: Optional[int] = None
    report_count_above: Optional[int] = None
    report_count_below: Optional[int] = None

